//
//  AppDelegate.swift
//  Dictionary
//
//  Created by Phidgets on 2018-02-27.
//  Copyright © 2018 Phidgets. All rights reserved.
//

import Cocoa
import Phidget22Swift

@NSApplicationMain
class AppDelegate: NSObject, NSApplicationDelegate, NSTableViewDelegate, NSTableViewDataSource {

    @IBOutlet weak var window: NSWindow!
    let ch = Dictionary()
    let phidgetInfoBox = PhidgetInfoBox()
    
    //GUI
    @IBOutlet weak var scrollView: NSScrollView!
    @IBOutlet weak var tableView: NSTableView!
    @IBOutlet weak var dictionaryLabel: NSTextField!
    @IBOutlet weak var serialNumberLabel: NSTextField!
    @IBOutlet weak var keyText: NSTextField!
    @IBOutlet weak var valueText: NSTextField!
    
    //Error Window
    @IBOutlet weak var errorEventWindow: NSPanel!
    @IBOutlet var errorEventLog: NSTextView!
    @IBOutlet weak var errorEventCount: NSTextField!
    var errorCounter = 0

    //Misc
    var pairs = [String : String]()
    var sortedKeys = [String]()
    
    //MARK: View events
    func applicationDidFinishLaunching(_ aNotification: Notification) {
        do {
            //Adjusting window height
            var frame = self.window.frame
            frame.size.height = self.window.minSize.height
            self.window.setFrame(frame, display: true, animate: false)
            self.window.center()
    
            let _ = ch.attach.addHandler(attach_handler)
            let _ = ch.detach.addHandler(detach_handler)
            let _ = ch.error.addHandler(error_handler)
            let _ = ch.add.addHandler(add_handler)
            let _ = ch.update.addHandler(update_handler)
            let _ = ch.remove.addHandler(remove_handler)
            
            /*
             * Please review the Phidget22 channel matching documentation for details on the device
             * and class architecture of Phidget22, and how channels are matched to device features.
             */
            
            /*
             * Specifies the serial number of the device to attach to.
             * For VINT devices, this is the hub serial number.
             *
             * The default is any device.
             */
            // try ch.setDeviceSerialNumber(<YOUR DEVICE SERIAL NUMBER>)
            
            /*
             * In order to attach to a network Phidget, the program must connect to a Phidget22 Network Server.
             * In a normal environment this can be done automatically by enabling server discovery, which
             * will cause the client to discovery and connect to available servers.
             *
             * To force the channel to only match a network Phidget, set remote to 1.
             */
            
            // try Net.enableServerDiscovery(serverType: .deviceRemote)
            // try ch.setIsRemote(true)
            
            /*
             openCmdLine allows command line arguments to be passed to this program in order to
             configure channel matching.
             If you have modified the code and added your own channel matching, comment out
             the openCmdLine call, and simply replace it with the following line:
             */
            //try ch.open()
            
            try phidgetInfoBox.openCmdLine(phid: ch as Phidget)
            
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    func applicationShouldTerminateAfterLastWindowClosed(_ sender: NSApplication) -> Bool {
        return true
    }
    
    func applicationWillTerminate(_ aNotification: Notification) {
        //Safely close example
        do{
            print("Closing")
            ch.attach.removeAllHandlers()
            ch.detach.removeAllHandlers()
            ch.error.removeAllHandlers()
            ch.add.removeAllHandlers()
            ch.update.removeAllHandlers()
            ch.remove.removeAllHandlers()
            try ch.close()
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    //MARK: Event handlers
    func attach_handler(sender: Phidget){
        do{
            let scanner = DictionaryScanner()
            let attachedDevice = sender as! Phidget22Swift.Dictionary
            let serialNumber = try attachedDevice.getDeviceSerialNumber()
            let name = try attachedDevice.getDeviceLabel()
            pairs = (try scanner.scan(dictionary: attachedDevice))
            sortAndUpdate()
            DispatchQueue.main.async {
                self.serialNumberLabel.stringValue = String(format:"%d", serialNumber)
                self.dictionaryLabel.stringValue = name
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    func detach_handler(sender: Phidget){
        DispatchQueue.main.async {
            //Adjusting window height
            var frame = self.window.frame
            frame.size.height = self.window.minSize.height
            self.window.setFrame(frame, display: true, animate: false)
            self.window.center()
        }
    }
    
    func error_handler(sender: Phidget, data: (code: ErrorEventCode, description: String)){
        outputError(errorDescription: data.description, errorCode: data.code.rawValue)
    }
    
    func add_handler(sender: Phidget22Swift.Dictionary, data: (key: String, value: String)){
        pairs.updateValue(data.value, forKey: data.key)
        sortAndUpdate()
    }
    
    func remove_handler(sender: Phidget22Swift.Dictionary, key: String){
        pairs.removeValue(forKey: key)
        sortAndUpdate()
        
    }
    
    func update_handler(sender: Phidget22Swift.Dictionary, data: (key: String, value: String)){
        pairs.updateValue(data.value, forKey: data.key)
        sortAndUpdate()
    }
    
    //MARK: GUI Controls
    @IBAction func addEntry(_ sender: Any) {
        do{
            try ch.add(key: keyText.stringValue, value: valueText.stringValue)
            pairs.updateValue(valueText.stringValue, forKey: keyText.stringValue)
            sortAndUpdate()
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func removeEntry(_ sender: Any) {
        do{
            if(tableView.selectedRow == -1){
                return
            }
            try ch.remove(key: sortedKeys[tableView.selectedRow])
            pairs.removeValue(forKey: sortedKeys[tableView.selectedRow])
            sortAndUpdate()
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    //MARK: Misc
    func sortAndUpdate(){
        sortedKeys.removeAll()
        sortedKeys = pairs.keys.sorted()
        DispatchQueue.main.async {
            self.tableView.reloadData()
        }
    }

    //MARK: Table View
    func numberOfRows(in tableView: NSTableView) -> Int {
        return pairs.count
    }
    
    func tableView(_ tableView: NSTableView, viewFor tableColumn: NSTableColumn?, row: Int) -> NSView? {
        if((tableColumn?.identifier)!.rawValue == "keyColumn"){
            let result = tableView.makeView(withIdentifier: NSUserInterfaceItemIdentifier.init("keyView"), owner: nil) as! NSTableCellView
            if(pairs.count == 0){
                result.textField?.stringValue = ""
            }
            else{
                result.textField?.stringValue = sortedKeys[row]
            }
            return result
        }
        else{
            let result = tableView.makeView(withIdentifier: NSUserInterfaceItemIdentifier.init("valueView"), owner: nil) as! NSTableCellView
            if(pairs.count == 0){
                result.textField?.stringValue = ""
            }
            else{
                result.textField?.stringValue = pairs[sortedKeys[row]]!
            }
            return result
        }
    }
    
    //MARK: Error
    func outputError(errorDescription:String, errorCode:UInt32){
        DispatchQueue.main.async {
            self.errorCounter += 1
            let outputString = NSAttributedString(string: String(format:"Error %d: %@\n", errorCode, errorDescription))
            self.errorEventLog.textStorage?.beginEditing()
            self.errorEventLog.textStorage?.append(outputString)
            self.errorEventLog.textStorage?.endEditing()
            self.errorEventCount.stringValue = "\(self.errorCounter)"
            if(!self.errorEventWindow.isVisible){
                self.errorEventWindow.setIsVisible(true)
                self.errorEventWindow.center()
            }
        }
        
    }
    
    @IBAction func clearError(_ sender: Any) {
        errorEventLog.textStorage?.setAttributedString(NSAttributedString(string: ""))
        errorCounter = 0
        errorEventCount.stringValue = "0"
    }
}

